from enum import Enum

from .condition import Condition
from .base_conditions import and_, msg, or_, msgtype

# TODO: Add tests
# TODO: Add other fields that are common for logs

_is_foxglove = or_(msgtype == "foxglove_msgs/Log", msgtype == "foxglove.Log")
_is_ros = msgtype == "rosgraph_msgs/Log"

LogLevel = Enum("LogLevel", ["UNKNOWN", "DEBUG", "INFO", "WARN", "ERROR", "FATAL"])

log = or_(
    and_(_is_ros, msg.msg),
    and_(_is_foxglove, msg.message),
    # Default case
    "",
)

log_level = or_(
    and_(_is_ros, Condition.map(msg, lambda m: ros_log_level(m.level))),
    and_(_is_foxglove, Condition.map(msg, lambda m: foxglove_log_level(m.level))),
    # Default case
    LogLevel.UNKNOWN,
)


def ros_log_level(num):
    if num == 1:
        return LogLevel.DEBUG
    elif num == 2:
        return LogLevel.INFO
    elif num == 4:
        return LogLevel.WARN
    elif num == 8:
        return LogLevel.ERR
    elif num == 16:
        return LogLevel.FATAL
    else:
        return LogLevel.UNKNOWN


def foxglove_log_level(num):
    if num == 1:
        return LogLevel.DEBUG
    elif num == 2:
        return LogLevel.INFO
    elif num == 3:
        return LogLevel.WARN
    elif num == 4:
        return LogLevel.ERROR
    elif num == 5:
        return LogLevel.FATAL
    else:
        return LogLevel.UNKNOWN


__all__ = [
    "log",
    "log_level",
    "LogLevel",
]



import unittest

from ruleengine.dsl.condition import Condition


class ConditionTest(unittest.TestCase):
    def test_arithmetics(self):
        self.assertEqual(-4, int(self._get_eval_result(-Condition.wrap(4))))
        self.assertEqual(4, int(self._get_eval_result(+Condition.wrap(4))))

        self.assertEqual(10, int(self._get_eval_result(Condition.wrap(4) + 6)))
        self.assertEqual(10, int(self._get_eval_result(6 + Condition.wrap(4))))

        self.assertEqual(-2, int(self._get_eval_result(Condition.wrap(4) - 6)))
        self.assertEqual(2, int(self._get_eval_result(6 - Condition.wrap(4))))

        self.assertEqual(24, int(self._get_eval_result(Condition.wrap(4) * 6)))
        self.assertEqual(24, int(self._get_eval_result(6 * Condition.wrap(4))))

        self.assertEqual(2, int(self._get_eval_result(Condition.wrap(12) / 6)))
        self.assertEqual(2, int(self._get_eval_result(12 / Condition.wrap(6))))

    def _get_eval_result(self, cond):
        return cond.evaluate_condition_at(None, {})[0]



import re

from .condition import Condition, ThunkCondition

always = Condition.wrap(True)
msg = ThunkCondition(lambda item, scope: (item.msg, scope))
ts = ThunkCondition(lambda item, scope: (item.ts, scope))
topic = ThunkCondition(lambda item, scope: (item.topic, scope))
msgtype = ThunkCondition(lambda item, scope: (item.msgtype, scope))


def get_start_time(item, scope):
    value = item.ts
    if "start_time" in scope:
        value = scope["start_time"]
    return value, scope


condition_start_time = ThunkCondition(get_start_time)


@Condition.wrap_args
def and_(*conditions):
    assert len(conditions) > 0, "and_ must have at least 1 condition"

    def new_thunk(item, scope):
        value = True
        for condition in conditions:
            value, scope = condition.evaluate_condition_at(item, scope)
            if not value:
                break
        return value, scope

    return ThunkCondition(new_thunk)


@Condition.wrap_args
def or_(*conditions):
    assert len(conditions) > 0, "or_ must have at least 1 condition"

    def new_thunk(item, scope):
        value = False
        for condition in conditions:
            value, scope = condition.evaluate_condition_at(item, scope)
            if value:
                break
        return value, scope

    return ThunkCondition(new_thunk)


@Condition.wrap_args
def not_(condition):
    return Condition.map(condition, lambda x: not x)


@Condition.wrap_args
def is_none(condition):
    def new_thunk(item, scope):
        value, scope = condition.evaluate_condition_at(item, scope)
        return value is None, scope

    return ThunkCondition(new_thunk)


@Condition.wrap_args
def concat(*pieces):
    def new_thunk(item, scope):
        str_pieces = []
        for p in pieces:
            value, scope = p.evaluate_condition_at(item, scope)
            str_pieces.append(str(value))
        return "".join(str_pieces), scope

    return ThunkCondition(new_thunk)


@Condition.wrap_args
def get_value(key):
    return Condition.flatmap(
        key, lambda k: ThunkCondition(lambda item, scope: (scope[k], scope))
    )


@Condition.wrap_args
def set_value(key, value):
    return Condition.apply(
        lambda scope, actual_key, actual_value: (
            True,
            {**scope, actual_key: actual_value},
        ),
        key,
        value,
    )


@Condition.wrap_args
def has(parent, child):
    return Condition.apply(
        lambda scope, p, c: (c in p, {**scope, "cos/contains": c if c in p else None}),
        parent,
        child,
    )


def regex(value, pattern):
    return Condition.flatmap(
        Condition.map(Condition.wrap(value), lambda v: re.search(pattern, v)),
        lambda match_result: ThunkCondition(
            lambda item, scope: (match_result, {**scope, "cos/regex": match_result})
        ),
    )


def func_apply(func, *args):
    return Condition.apply(
        lambda scope, f, *a: (f(*a), scope),
        Condition.wrap(func),
        *[Condition.wrap(arg) for arg in args]
    )


__all__ = [
    "always",
    "msg",
    "ts",
    "topic",
    "msgtype",
    "condition_start_time",
    "and_",
    "or_",
    "not_",
    "concat",
    "get_value",
    "set_value",
    "has",
    "regex",
    "is_none",
    "func_apply",
]



import ast
from ruleengine.dsl.validation.normalizer import normalize_expression_tree
from ruleengine.dsl.validation.validator import base_dsl_values


def str_to_condition(expr_str, additional_injected_values=None):
    parsed = ast.parse(expr_str, mode="eval")
    code = compile(normalize_expression_tree(parsed), "", mode="eval")
    return eval(code, {**base_dsl_values, **(additional_injected_values or {})})



import ast


def normalize_expression_tree(tree):
    return ast.fix_missing_locations(BooleanTransformer().visit(tree))


class BooleanTransformer(ast.NodeTransformer):
    def visit_JoinedStr(self, node):
        new_values = []
        condition_args = []
        for v in node.values:
            if isinstance(v, ast.Constant):
                new_values.append(v)
            elif isinstance(v, ast.FormattedValue):
                if v.format_spec:
                    # In theory, f-strings format spec can be another
                    # f-string, but it makes our logic very clumsy, so we
                    # just don't support it for now. Check here that it's
                    # only a constant.
                    if len(v.format_spec.values) != 1 or not isinstance(
                        v.format_spec.values[0], ast.Constant
                    ):
                        raise Exception("Formatting does not support nested values")

                new_values.append(
                    ast.FormattedValue(
                        ast.Name(f"arg{len(condition_args)}", ast.Load()),
                        v.conversion,
                        v.format_spec,
                    )
                )
                condition_args.append(self.generic_visit(v.value))
            else:
                raise Exception(f"Shouldn't get here: {v}")

        lambda_node = self._eval_expr(
            f"lambda {','.join(f'arg{i}' for i in range(len(condition_args)))}: ..."
        )
        lambda_node.body = ast.JoinedStr(new_values)
        return ast.Call(
            ast.Name("func_apply", ast.Load()), [lambda_node] + condition_args, []
        )

    def visit_BoolOp(self, node):
        node = self.generic_visit(node)
        if isinstance(node.op, ast.And):
            func_name = "and_"
        elif isinstance(node.op, ast.Or):
            func_name = "or_"
        else:
            return node
        return ast.Call(ast.Name(func_name, ast.Load()), node.values, [])

    def visit_UnaryOp(self, node):
        node = self.generic_visit(node)
        if isinstance(node.op, ast.Not):
            return ast.Call(ast.Name("not_", ast.Load()), [node.operand], [])
        else:
            return node

    def visit_Call(self, node):
        # A hack since 1. conditions have states, and 2. within SequenceMatchCondition,
        # we need to create clean copies of the condition objects to avoid state
        # sharing. As a result, we need to transform the condition arguments
        # of the SequenceMatchCondition related functions to condition factories,
        # with which we can lazy create copies of the condition objects
        # within the SequenceMatchCondition.
        node = self.generic_visit(node)
        if isinstance(node.func, ast.Name) and node.func.id in ("repeated", "debounce"):
            factory = self._eval_expr("lambda: ...")
            factory.body = node.args[0]
            node.args[0] = factory
        elif isinstance(node.func, ast.Name) and node.func.id in (
            "sequential",
            "timeout",
        ):
            new_args = []
            for arg in node.args:
                factory = self._eval_expr("lambda: ...")
                factory.body = arg
                new_args.append(factory)
            node.args = new_args
        return node

    def visit_Compare(self, node):
        # We need to jump through quite a few hoops to keep inline with Python's
        # semantics for comparison operators, just so we can rewrite the `in`
        # operator into our `has` function call. Suppose you have the expression
        #
        #  a op1 b op2 c
        #
        # Python treats this as
        #
        #  a op1 b and b op2 c
        #
        # Except `b` must only be evaluated once. So we package the above into
        # the following:
        #
        # (lambda arg0, arg1, arg2:
        #   and_(
        #     arg0 op1 arg1,
        #     arg1 op2 arg2))(a, b, c)
        #
        # And if one of the operators happen to be an `in` operator, we swap it
        # out with a `has` call.

        node = self.generic_visit(node)
        args = [node.left] + node.comparators
        param_list = ["arg" + str(i) for i in range(len(args))]
        condition_list = []
        for i, (left, right) in enumerate(zip(param_list, param_list[1:])):
            if isinstance(node.ops[i], ast.In):
                condition_list.append(self._eval_expr(f"has({right}, {left})"))
            elif isinstance(node.ops[i], ast.NotIn):
                condition_list.append(self._eval_expr(f"not_(has({right}, {left}))"))
            else:
                condition_list.append(
                    ast.Compare(
                        ast.Name(left, ast.Load()),
                        [node.ops[i]],
                        [ast.Name(right, ast.Load())],
                    )
                )

        wrapper = self._eval_expr(f"lambda {','.join(param_list)}: ...")
        wrapper.body = ast.Call(ast.Name("and_", ast.Load()), condition_list, [])
        return ast.Call(wrapper, args, [])

    def _eval_expr(self, str_value):
        return ast.parse(str_value, mode="eval").body



import logging
import os
import re
from semver.version import Version

from setuptools import setup, find_packages

version = os.environ.get("RULE_ENGINE_VERSION")
if not version:
    logging.info("RULE_ENGINE_VERSION not set, might be running in CI")
    version = "0.0.0"

# Remove the leading "v" if present
version = re.sub("^v", "", version)
# Semantically check the version
version = str(Version.parse(version))


def requires(filename="requirements.txt"):
    with open(filename) as f:
        return f.read().splitlines()


setup(
    name="cos-ruleengine",
    version=version,
    description="",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=requires(),
    extras_require={"dev": requires("requirements-dev.txt")},
    python_requires=">=3.7",
)



from sys import argv
import json
from ruleengine.dsl.base_actions import noop
from .config_validator import validate_config

if __name__ == "__main__":
    config = json.loads(argv[1])
    result, _ = validate_config(config, noop)

    print(json.dumps(result))
    exit(0 if result["success"] else 1)
